<?php

// "Baked" template is just a list of strings & slot references
$baked_template = null;
$baking = false;

// Start baking a template.
function template_begin() {
	global $baking, $baked_template;

	if ($baking)
		throw new Exception('template_begin() called during baking');
	if ($baked_template != null)
		throw new Exception('template_begin() called after template was baked');

	$baking = true;
	$baked_template = array();

	if (!ob_start('on_template_flush'))
		throw new Exception('failed to start output buffer');
}

$next_flush_slot = null;

function on_template_flush(string $buffer) {
	global $baked_template, $next_flush_slot;

	$baked_template[] = array('content' => $buffer, 'slot' => $next_flush_slot);
	$next_flush_slot = null;

	return '';
}

// Add a slot to the template being baked.
function bake_slot(string $slot) {
	global $baking, $next_flush_slot;

	if (!$baking)
		throw new Exception('bake_slot() called outside of baking');

	$next_flush_slot = $slot;
	ob_clean();
}

// Finish baking the template.
function template_end() {
	global $baking, $baked_template;

	if (!$baking)
		throw new Exception('template_end() called outside of baking');

	ob_end_clean();
	$baking = false;
}

$slots = array();

// Output template content until the next missing slot.
function to_next_missing_slot(): ?string {
	global $slots, $baked_template;

	if ($baked_template == null)
		throw new Exception('cannot output, no baked template');

	while (count($baked_template) > 0) {
		$next = array_shift($baked_template);
		echo $next['content'];

		if ($next['slot'] == null)
			continue;

		if (array_key_exists($next['slot'], $slots)) {
			echo $slots[$next['slot']];
		} else {
			array_unshift($baked_template, array('content' => '', 'slot' => $next['slot']));
			return $next['slot'];
		}
	}

	return null;
}

$buffering_slot = null;
function slot_begin(string $slot) {
	global $buffering_slot, $baked_template;

	$next_slot = to_next_missing_slot();

	if ($next_slot != $slot) {
		if (!ob_start())
			throw new Exception('failed to start output buffer');
		$buffering_slot = $slot;
	} else {
		array_shift($baked_template);
	}
}

function slot_end() {
	global $buffering_slot, $slots;

	if ($buffering_slot != null) {
		$content = ob_get_clean();
		$slots[$buffering_slot] = $content;
		$buffering_slot = null;
	}

	to_next_missing_slot();
}

function slot_set_empty(string $slot) {
	global $slots;
	$slots[$slot] = '';
}

function slot_unset(string $slot) {
	global $slots;
	unset($slots[$slot]);
}

function mail_exception_trace(Throwable $ex): string {
	$output = "";
	while ($ex != null) {
		$output .= get_class($ex) . ": " . $ex->getMessage() . "\r\n";
		$output .= "-- " . $ex->getFile() . "(" . $ex->getLine() . ") Code=" . $ex->getCode() . "\r\n";
		$output .= $ex->getTraceAsString() . "\r\n";
		$ex = $ex->getPrevious();
	}
	return $output;
}

function html_exception_trace(Throwable $ex) {
	$count = 0;
	while ($ex != null) {
		echo "<blockquote>";
		echo "<p><b>" . htmlspecialchars(get_class($ex)) . ":</b> " . htmlspecialchars($ex->getMessage()) . "</p>";
		echo "<ul>";

		foreach (array_reverse($ex->getTrace()) as $trace) {
			$func = htmlspecialchars(($trace['class'] ?? '') . ($trace['type'] ?? '') . $trace['function']);
			$file = htmlspecialchars($trace['file'] ?? '?');
			$line = htmlspecialchars($trace['line'] ?? '?');
			echo "<li><code>$func()</code> called<br /><small>$file, line $line</small></li>";
		}

		$file = htmlspecialchars($ex->getFile());
		$line = htmlspecialchars($ex->getLine());
		echo "<li>Exception thrown<br /><small>$file, line $line</small></li></ul>";

		$ex = $ex->getPrevious();
		$count += 1;
	}
	for ($i = 0; $i < $count; $i++) {
		echo "</blockquote>";
	}
}

function display_exception(Throwable $ex) {
	$page_name = $_SERVER["PHP_SELF"] ?? "(unknown page)";

	$mail_body = "An uncaught exception was encountered on $page_name.\r\n";
	$mail_body .= "\r\nException details:\r\n";
	$mail_body .= "Type: " . get_class($ex) . ", Code: " . $ex->getCode() . ", Message: " . $ex->getMessage() . "\r\n";
	$mail_body .= "File: " . $ex->getFile() . ", Line: " . $ex->getLine() . "\r\n";
	$mail_body .= "\r\nTrace:\r\n";
	$mail_body .= mail_exception_trace($ex);
	$mail_body .= "\r\nRequest details:\r\n";
	$mail_body .= $_SERVER['REQUEST_METHOD'] . " " . $_SERVER['REQUEST_URI'] . "\r\n";
	foreach (getallheaders() as $name => $value) {
		$mail_body .= "$name: $value\r\n";
	}
	foreach ($_REQUEST as $key => $value) {
		$mail_body .= "\$_REQUEST[$key]: $value\r\n";
	}

	mail("webmaster@keithhacks.cyou", "Exception on $page_name", $mail_body);
	for ($i = ob_get_level(); $i > 0; $i--) {
		ob_end_flush();
	}
	if (!headers_sent())
		http_response_code(500);

	$display_trace = true;
	?>
	<div class="uncaught-exception">
		<img src="/assets/sorry_nothing.png" role="presentation" />
		<h1>Internal Server Error</h1>
		<p>An uncaught exception was encountered on this page.</p>
		<p><b><?=get_class($ex)?>:</b> <?=$ex->getMessage()?></p>
		<p>Please contact <a href="mailto:webmaster@keithhacks.cyou">webmaster@keithhacks.cyou</a> if this problem persists.</p>
		<?php
		if ($display_trace) {
			echo "<h2>Exception details</h2>";
			html_exception_trace($ex);
		}
		?>
	</div>
	<?php
}
set_exception_handler('display_exception');
?>
